/*
    Sequences cuts container.
    Copyright (C) 2021  Alessandro Accattini

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 *****************************************************************************
 *
 *
 * $Header$
 *
 *
 *
 * REVISION HISTORY:
 * $Log$
 */


/***************************************************************************/
/* include vari */

#include "magiceye/magiceye.h"			// magic-eye APIs
#include "cuts.h"

/***************************************************************************/
/* define interni alla libreria */

/***************************************************************************/
/* strutture contenute all'interno della libreria */

typedef struct
{
  int		Pos;		// posizione finale, ingresso in Fluid
  unsigned int	Duration;	// durata dello stimolo
}Move_t;

/***************************************************************************/
/* funzioni interne alla libreria */

static unsigned int MoveRing(void *Data,unsigned int Call);
  /* Muove il ring secondo una tabella pre-impostata. */


static unsigned int Ring_2_Step_01(void *Data,unsigned int Call);
static unsigned int Ring_2_Step_02(void *Data,unsigned int Call);
static unsigned int Ring_2_Step_03(void *Data,unsigned int Call);
static unsigned int Ring_2_Step_04(void *Data,unsigned int Call);
static unsigned int Ring_2_Step_05(void *Data,unsigned int Call);
static unsigned int Ring_2_Step_06(void *Data,unsigned int Call);

/***************************************************************************/
/* costanti interne alla libreria */

static const Move_t	MoveList[:>={	{ 100,80},
					{  50,50},
					{   0,40},
					{  30,45},
					{ 130,80},
					{ 100,40},
					{ 130,60},
					{ 160,60},
					{ 110,60},
					{  80,50},
					{   0,70},
					{   0, 0}};	// end of list


static const SeqCut_t		SeqData_1[:>={
					MoveRing,
					NULL};

static const SeqCut_t		SeqData_2[:>={
					Ring_2_Step_01,	// rot. DX
					Ring_2_Step_02,	// close
					Ring_2_Step_03,	// pause
					Ring_2_Step_04,	// open
					Ring_2_Step_05,	// pause
					Ring_2_Step_06,	// rot. SX
					Ring_2_Step_02,	// close
					Ring_2_Step_03,	// pause
					Ring_2_Step_04,	// open
					Ring_2_Step_05,	// pause
					NULL};

/***************************************************************************/
/* variabili interne alla libreria */

/***************************************************************************/
/* variabili pubbliche */

CREATE_SEQUENCER(Seq_1,SeqData_1);		// public sequencer
CREATE_SEQUENCER(Seq_2,SeqData_2);		// public sequencer

unsigned char		FluidDebug=0;		// debug flag

/***************************************************************************/
/* funzioni */

static unsigned int MoveRing(void *Data,unsigned int Call)
{
  /* Muove il ring secondo una tabella pre-impostata. */



  static const Move_t	*Entry=MoveList;	// puntatore alla lista
  static unsigned int	Line=0;			// # riga nel fluid log
  SeqData_t		*SeqData=(SeqData_t*)Data;
  Fluid_IO_t		Rot_1,Rot_2;


  
  // produce l'uscita di Fluid
  Rot_1=	Fluid_Move(&SeqData->A.FluidFilt,Entry->Pos);
  Rot_2=	Fluid_Move(&SeqData->B.FluidFilt,Rot_1);
  SeqData->A.Rotate[0]=	Rot_1;
  SeqData->B.Rotate[0]=	Rot_2;
  SeqData->B.Rotate[1]=	Rot_2;

  if(FluidDebug)				// se abilitato
    printf("%u\t%u\t%hi\t%hi\t%hi\n",Line++,Call,Entry->Pos,Rot_1,Rot_2);
  
  // verifica se lo stimilo e' finito
  if((Call + 1) >= Entry->Duration)		// se la fase e' conclusa
  {
    // inizia una nuova fase del movimento
    Entry++;					// next entry
    if(Entry->Duration == 0)			// se e' la fine della lista
    {
      Entry=MoveList;				// riparte da capo
      if(FluidDebug)
	puts("# end of list reached");
    }
    if(FluidDebug)
      puts("# new Fluid input...");
    Fluid_Start(&SeqData->A.FluidFilt,Entry->Duration,Entry->Pos);
    return 0;					// cut terminato
  }
  return 1;					// il cut prosegue
}





static unsigned int Ring_2_Step_01(void *Data,unsigned int Call)
{
  /* Ruota verso DX di 16 passi. */



  SeqData_2_t	*SeqData=(SeqData_2_t*)Data;



  if(Call >= 16)				// se ha finito
    return 0;					// fine sequenza
  SeqData->Rotate[0]++;				// ruota verso DX
  SeqData->Rotate[1]--;				// ruota verso SX
  return 1;					// ancora attivo
}



static unsigned int Ring_2_Step_02(void *Data,unsigned int Call)
{
  /* Stringe i segmenti nel primo anello. */
  
  
  
  SeqData_2_t	*SeqData=(SeqData_2_t*)Data;
  Color_t	Color;				// pixel color
  int		i;



  if(Call >= 6)					// se ha finito
    return 0;					// fine sequenza
  for(i=0;i<16;i++)				// pixel di 1/4 di giro
  {
    // primo/terzo quarto
    Color=(((i+(int)Call) >= 7) && ((i+(int)Call) <= 8)) ? 0xffff : 0x0000;
    MagicEye_PutPixel(SeqData->Eye,i   ,0,Color);
    MagicEye_PutPixel(SeqData->Eye,i+32,0,Color);

    // secondo/quarto quarto
    Color=(((i-(int)Call) >= 7) && ((i-(int)Call) <= 8)) ? 0xffff : 0x0000;
    MagicEye_PutPixel(SeqData->Eye,i+16   ,0,Color);
    MagicEye_PutPixel(SeqData->Eye,i+16+32,0,Color);
  }
  return 1;					// ancora attivo
}



static unsigned int Ring_2_Step_03(void *Data,unsigned int Call)
{
  /* Pausa. */
  
  
  
  if(Call >= 10)				// se ha finito
    return 0;					// fine sequenza
  return 1;					// ancora attivo
}



static unsigned int Ring_2_Step_04(void *Data,unsigned int Call)
{
  /* Allarga i segmenti nel primo anello. */
  
  
  
  SeqData_2_t	*SeqData=(SeqData_2_t*)Data;
  Color_t	Color;				// pixel color
  int		i;



  if(Call >= 6)					// se ha finito
    return 0;					// fine sequenza
  for(i=0;i<16;i++)				// pixel di 1/4 di giro
  {

    // primo/terzo quarto
    Color=(((i-(int)Call) >= 2) && ((i-(int)Call) <= 3)) ? 0xffff : 0x0000;
    MagicEye_PutPixel(SeqData->Eye,i   ,0,Color);
    MagicEye_PutPixel(SeqData->Eye,i+32,0,Color);

    // secondo/quarto quarto
    Color=(((i+(int)Call) >= 12) && ((i+(int)Call) <= 13)) ? 0xffff : 0x0000;
    MagicEye_PutPixel(SeqData->Eye,i+16   ,0,Color);
    MagicEye_PutPixel(SeqData->Eye,i+16+32,0,Color);
  }
  return 1;					// ancora attivo
}



static unsigned int Ring_2_Step_05(void *Data,unsigned int Call)
{
  /* Pausa. */
  
  
  
  if(Call >= 5)					// se ha finito
    return 0;					// fine sequenza
  return 1;					// ancora attivo
}



static unsigned int Ring_2_Step_06(void *Data,unsigned int Call)
{
  /* Ruota verso SX di 32 passi. */



  SeqData_2_t	*SeqData=(SeqData_2_t*)Data;



  if(Call >= 32)				// se ha finito
    return 0;					// fine sequenza
  SeqData->Rotate[0]--;				// ruota verso DX
  SeqData->Rotate[1]++;				// ruota verso SX
  return 1;					// ancora attivo
}

/* $Source$ */

