/*
    Magic eye demo.
    Copyright (C) 2017 2018 2021  Alessandro Accattini

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 *****************************************************************************
 *
 *
 * $Header$
 *
 *
 *
 * REVISION HISTORY:
 * $Log$
 */


/***************************************************************************/
/* include vari */

#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include "magiceye/magiceye.h"			// magic eye control
#include "loadpng.h"				// png helper
#include "random.h"				// random generator
#include "fluid/fluid.h"			// fluid movement filter
#include "stretch.h"				// image stretching
#include "cuts.h"				// sequences, cuts, etc...

#define ENABLE_DEBUG		1		// enabled if 1
#include "macros.h"				// debug macro


extern MagicEye_t	Eye_1;			// magic eye descriptor
extern MagicEye_t	Eye_2;			// magic eye descriptor
extern MagicEye_t	Eye_3;			// magic eye descriptor

/***************************************************************************/
/* define vari */

#define ENABLE_PIXMAP_VIEW	0		// enabled if !0


#define UPDATE_TIME		50		// refresh ogni 50ms
#define VIDEO_W			320
#define VIDEO_H			240
#define VIDEO_MODE		(SDL_HWSURFACE|SDL_DOUBLEBUF)
#define VSCALE			6

/***************************************************************************/
/* strutture */

typedef struct
{
  int		Pos;		// posizione finale
  int		Steps;		// durata del movimento
}Move_t;

/***************************************************************************/
/* prototipi funzioni */

static void PrintMagicEyeData(const char *EyeName,const MagicEye_t *EyeDesc);
  /* Stampa alcuni dati relativi al magic eye. */

static Uint32 TimerCallback(Uint32 Interval,void *Param);
  /* Funzione di callback del timer, esegue la lettura dei registri nel
     device remoto. */

/***************************************************************************/
/* costanti */

static const char	Licenza[:>=
	"demo version 0.05, Copyright (C) 2017 2018 2021  "
	"Alessandro Accattini\n";
static const char	Help[:>=
	"Usage: demo [OPTIONS]\n"
	"\t-a ALPHA\tset the magic eye alpha level (default=255).\n"
	"\t-h\tdisplays this help.\n"
	"\t-d\tenables fluid data debug.\n"
	"\t-t TIME\tchange the refresh interval (default=50).\n"
	"\nReport bugs to: <bit.trapper@gmail.com>\n";

/***************************************************************************/
/* variabili */

static unsigned char	Run=1;			// run/exit flag
static SDL_Surface	*Video;			// finestra base
static SDL_Surface	*BackGround;		// sfondo
static SDL_Surface	*Magic_1;		// magic eye surface
static SDL_Surface	*Magic_2;		// magic eye surface
static SDL_Surface	*Magic_3;		// magic eye surface
#if ENABLE_PIXMAP_VIEW
static SDL_Surface	*Map;			// original linear pixels map
static SDL_Surface	*PixMap;		// stretched linear pixels map
#endif


static SeqData_t	SeqData_1={.A.Rotate={0,0,0}};	// local user data
static SeqData_2_t	SeqData_2={.Rotate={0,0}};	// local user data

/***************************************************************************/
/* funzioni */

static void PrintMagicEyeData(const char *EyeName,const MagicEye_t *EyeDesc)
{
  /* Stampa alcuni dati relativi al magic eye. */



  printf("%s:\n",EyeName);
  printf("  map array   : %i x %i\n",EyeDesc->u,EyeDesc->v);
  printf("  surface size: %i x %i\n",EyeDesc->w,EyeDesc->h);
  return;				// termina
}



static Uint32 TimerCallback(Uint32 Interval,void *Param)
{
  /* Funzione di callback del timer, esegue la lettura dei registri nel
     device remoto. */
  


  static SDL_Rect	Pos_1={.x=147-64,.y=120-64};
  static SDL_Rect	Pos_2={.x= 31-24,.y=143-24};
#if ENABLE_PIXMAP_VIEW
  static SDL_Rect	MapPos= {.x=  0   ,.y=  0   };
#endif



  // copia il background nella finestra finale
  SDL_BlitSurface(BackGround,NULL,Video,NULL);

  // anima il primo magic eye

  // primo anello
  Sequencer_Run(&Seq_1,&SeqData_1);		// uno step della sequenza

  // secondo anello
  SeqData_1.A.Rotate[1]--;			// ruota verso sinistra
  if(SeqData_1.A.Rotate[1] < 0)
    SeqData_1.A.Rotate[1]=127;
  RandomSector(SeqData_1.A.Eye,1);		// update random shape
  
  // terzo anello
  SeqData_1.A.Rotate[2]++;			// ruota verso destra
  if(SeqData_1.A.Rotate[2] >= 128)
    SeqData_1.A.Rotate[2]-=128;
  SeqData_1.B.Rotate[2]-=2;			// ruota verso sinistra
  if(SeqData_1.B.Rotate[2] < 0)
    SeqData_1.B.Rotate[2]+=128;
  
  // anima il secondo magic eye
  Sequencer_Run(&Seq_2,&SeqData_2);		// uno step della sequenza

 
  // aggiorna i magic eye
  MagicEye_Update(SeqData_1.A.Eye,SeqData_1.A.Rotate);
  MagicEye_Update(SeqData_1.B.Eye,SeqData_1.B.Rotate);
  MagicEye_Update(SeqData_2.Eye,  SeqData_2.Rotate);
  SDL_BlitSurface(Magic_1,NULL,Video,&Pos_1);		// magic eye
  SDL_BlitSurface(Magic_2,NULL,Video,&Pos_2);		// magic eye
  SDL_BlitSurface(Magic_3,NULL,Video,&Pos_1);		// magic eye
#if ENABLE_PIXMAP_VIEW
  SurfaceStretch( Map,PixMap,VSCALE);			// image stretch
  SDL_BlitSurface(PixMap,   NULL,Video,&MapPos );	// scaled eye map
#endif
  
  // aggiorna la finestra
  SDL_Flip(Video);					// window update
  return(Interval);					// termina
}





/***************************************************************************/

int main(int argc,char **argv)
{
  SDL_TimerID		SerialTimer;		// timer lettura registri
  SDL_Event		InputEvent;		// evento in ingresso
  unsigned long int	ReadInterval=UPDATE_TIME;
  int			Opzione;		// per la lettura delle opzioni
  int			i;
  unsigned char		Alpha=255;		// magic eye alpha level



  // visualizza la licenza del programma
  puts(Licenza);

  // legge le opzioni presenti sulla riga di comando
  while((Opzione=getopt(argc,argv,"hdt:a:"))!=-1)
  {
    switch(Opzione)
    {
      case 'h':
        puts(Help);				// stampa l'help
	return 0;				// termina
	break;
      case 'd':
        FluidDebug=1;				// abilita il debug di Fluid
	break;
      case 't':
        sscanf(optarg,"%li",&ReadInterval);
	break;
      case 'a':
        sscanf(optarg,"%hhu",&Alpha);
	break;
      default:
        puts("Opzione sconosciuta !!!");
	exit(0);				// termina
	break;
    }
  }

  // inizializza la SDL
  PDEBUG("%s: initializing SDL...\n",__FUNCTION__);
  SDL_putenv("SDL_VIDEO_X11_WMCLASS=littlegptracker");
  if(SDL_Init(SDL_INIT_VIDEO|SDL_INIT_TIMER)!=0)
  {
    /* termina causa errore di inizializzazione */
    fprintf(stderr,"SDL_Init: %s\n",SDL_GetError());
    return 1;					// termina con errore
  }
  atexit(SDL_Quit);				// registra la SDL_Quit()

  // crea la finestra base
  if((Video=SDL_SetVideoMode(VIDEO_W,VIDEO_H,16,VIDEO_MODE))==NULL)
  {
    // termina causa errore creazione finestra
    fprintf(stderr,"SDL_SetVideoMode: %s\n",SDL_GetError());
    return 1;					// termina con fallimento
  }
 
  // inizializza Fluid filter
  SeqData_1.A.Eye=	&Eye_1;
  SeqData_1.B.Eye=	&Eye_3;
  SeqData_2.Eye=	&Eye_2;

  Fluid_Create(&SeqData_1.A.FluidFilt,0.94f,7.0f,0);	// configura il filtro
  Fluid_Start( &SeqData_1.A.FluidFilt,2000,0);		// avvia il movimento
  Fluid_Create(&SeqData_1.B.FluidFilt,0.96f,2.0f,0);
  Fluid_Start( &SeqData_1.B.FluidFilt,2000,0);
  
  // carica tutte le immagini da disco
  LoadSetVideo(Video);					// reference win
  if((BackGround=Load16bitPng("eye_2.png")) == NULL)
    puts("Unable to load background");

  // crea il magic eye
  Magic_1=MagicEye_CreateSurfaces(SeqData_1.A.Eye);	// crea le surface
  SDL_SetColorKey(Magic_1,SDL_SRCCOLORKEY,0x0000);	// nero trasparente
  SDL_SetAlpha(Magic_1,SDL_SRCALPHA,Alpha);		// alpha level
  
  Magic_3=MagicEye_CreateSurfaces(SeqData_1.B.Eye);	// crea le surface
  SDL_SetColorKey(Magic_3,SDL_SRCCOLORKEY,0x0000);	// nero trasparente
  SDL_SetAlpha(Magic_3,SDL_SRCALPHA,Alpha-50);		// alpha level

#if ENABLE_PIXMAP_VIEW
  Map=MagicEye_GetPixmap(SeqData_1.A.Eye);
  PixMap=LoadNewSurface(Map->w,Map->h * VSCALE);
#endif

  Magic_2=MagicEye_CreateSurfaces(SeqData_2.Eye);	// crea le surface
  SDL_SetColorKey(Magic_2,SDL_SRCCOLORKEY,0x0000);	// nero trasparente
  SDL_SetAlpha(Magic_2,SDL_SRCALPHA,Alpha);		// alpha level
  
  // colora il primo magic eye
  MagicEye_Fill(SeqData_1.A.Eye,0x0000);		// tutto nero
  MagicEye_Fill(SeqData_1.B.Eye,0x0000);
  
  // primo anello
  for(i=0;i<32;i++)
    MagicEye_PutPixel(SeqData_1.A.Eye,i,0,0xffff);	// init. del bianco
  for(i=64;i<80;i++)
    MagicEye_PutPixel(SeqData_1.A.Eye,i,0,0xffff);	// init. del bianco
  
  // secondo anello
  for(i=32;i<40;i++)
    MagicEye_PutPixel(SeqData_1.A.Eye,i,1,0xffff);	// init. del bianco
  for(i=120;i<128;i++)
    MagicEye_PutPixel(SeqData_1.A.Eye,i,1,0xffff);	// init. del bianco
  RandomSector(SeqData_1.A.Eye,1);			// random shape
  
  // terzo anello
  for(i=0;i<26;i++)
    MagicEye_PutPixel(SeqData_1.A.Eye,i,2,0xffff);	// init. del bianco
  for(i=72;i<88;i++)
    MagicEye_PutPixel(SeqData_1.A.Eye,i,2,0xffff);	// init. del bianco
  
  // colora il secondo magic eye
  MagicEye_Fill(SeqData_2.Eye,0x0000);			// tutto nero

  MagicEye_PutPixel(SeqData_2.Eye, 7   ,0,0xffff);	// pixel bianchi
  MagicEye_PutPixel(SeqData_2.Eye, 8   ,0,0xffff);
  MagicEye_PutPixel(SeqData_2.Eye, 7+16,0,0xffff);
  MagicEye_PutPixel(SeqData_2.Eye, 8+16,0,0xffff);
  MagicEye_PutPixel(SeqData_2.Eye, 7+32,0,0xffff);
  MagicEye_PutPixel(SeqData_2.Eye, 8+32,0,0xffff);
  MagicEye_PutPixel(SeqData_2.Eye, 7+48,0,0xffff);
  MagicEye_PutPixel(SeqData_2.Eye, 8+48,0,0xffff);
  for(i=0;i<11;i++)
  {
    MagicEye_PutPixel(SeqData_2.Eye,i+0 +3,1,0xffff);	// pixel bianco
    MagicEye_PutPixel(SeqData_2.Eye,i+16+3,1,0xffff);
    MagicEye_PutPixel(SeqData_2.Eye,i+32+3,1,0xffff);
    MagicEye_PutPixel(SeqData_2.Eye,i+48+3,1,0xffff);
  }
  
  // colora il terzo magic eye
  for(i=0;i<75;i++)
    MagicEye_PutPixel(SeqData_1.B.Eye,i,0,0x9fff);	// init. azzurro
  for(i=0;i<96;i++)
    MagicEye_PutPixel(SeqData_1.B.Eye,i,1,0x9fff);	// init. azzurro
  for(i=0;i<70;i++)
    MagicEye_PutPixel(SeqData_1.B.Eye,i,2,0x9fff);	// init. azzurro

  // crea il timer aggiornamento video
  PDEBUG("%s: creating screen update timer...\n",__FUNCTION__);
  if((SerialTimer=SDL_AddTimer(ReadInterval,TimerCallback,NULL))==NULL)
    fprintf(stderr,"SDL_AddTimer: %s\n",SDL_GetError());

  // stampa alcuni dati sui magic eye
  PrintMagicEyeData("Eye_1",SeqData_1.A.Eye);
  PrintMagicEyeData("Eye_2",SeqData_2.Eye);
  
  // stampa le istruzioni
  puts("Press 'q' to exit.");



  // main loop
  while(Run)
  {
    usleep(20000);				// 20ms
    if(!SDL_PollEvent(&InputEvent))		// se non succede niente
      continue;					// fine del discorso

    switch(InputEvent.type)
    {
//      case SDL_MOUSEBUTTONDOWN:
//      case SDL_MOUSEBUTTONUP:
//      case SDL_MOUSEMOTION:
      case SDL_QUIT:
	Run=0;					// abbandona il loop
	break;
      case SDL_KEYDOWN:
        switch(InputEvent.key.keysym.sym)
	{
	  case SDLK_q:
	    Run=0;				// abbandona il loop
	    break;
/*	    
	  case SDLK_z:
	    Rotate_1[0]--;
	    printf("Rotate_1[0]=%i\n",Rotate_1[0]);
	    break;
	  case SDLK_x:
	    Rotate_1[0]++;
	    printf("Rotate_1[0]=%i\n",Rotate_1[0]);
	    break;
	  case SDLK_n:
	    Rotate_1[1]--;
	    printf("Rotate_1[1]=%i\n",Rotate_1[1]);
	    break;
	  case SDLK_m:
	    Rotate_1[1]++;
	    printf("Rotate_1[1]=%i\n",Rotate_1[1]);
	    break;
*/	    
	}
	break;
    }
  }

  // libera le risorse allocate
  puts("Exiting...");
  SDL_RemoveTimer(SerialTimer);			// rimuove il timer
  MagicEye_DestroySurfaces(SeqData_1.A.Eye);	// distrugge le surface
  MagicEye_DestroySurfaces(SeqData_1.B.Eye);
  MagicEye_DestroySurfaces(SeqData_2.Eye);
  SDL_FreeSurface(BackGround);			// dealloca il background
  SDL_FreeSurface(Video);			// dealloca il video
  return 0;					// termina
}

/* $Source$ */

