/*
    Fluid movement digital filter implementation.
    Copyright (C) 2017 2018 2021  Alessandro Accattini

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 *****************************************************************************
 *
 *
 * $Header$
 *
 *
 *
 * REVISION HISTORY:
 * $Log$
 */


/***************************************************************************/
/* include vari */

#include <stdio.h>				// printf()
#include <math.h>				// M_PI, etc.

#define __FLUID_C
#include "fluid.h"				// functions prototypes

/***************************************************************************/
/* define interni alla libreria */

#define ENABLE_DEBUG_MESSAGES		0	// enabled if !0


#define RADIX_POINT			16	// # bit decimali considerati

/***************************************************************************/
/* macro interne alla libreria */

#if !FLUID_USE_FLOATING_POINT
#define INT_TO_FIXED_POINT(Int)		(((Fluid_FixPt_t)Int)<<16)
#define FIXED_POINT_TO_INT(Fixed)	((Fluid_FixPt_t)(Fixed>>16))
#define FLOAT_TO_FIXED_POINT(Flt)	((Fluid_FixPt_t)((Flt)*(float)0x10000))
#else
#define INT_TO_FIXED_POINT(Int)		(float)(Int)
#define FIXED_POINT_TO_INT(Fixed)	(Fluid_IO_t)(Fixed)
#define FLOAT_TO_FIXED_POINT(Flt)	(Flt)
#endif

/***************************************************************************/
/* strutture contenute all'interno della libreria */

/***************************************************************************/
/* costanti pubbliche */

const char	FluidVersion[]="Fluid library version 0.02";
const char	FluidBuild[]=__DATE__;

/***************************************************************************/
/* variabili interne alla libreria */

/***************************************************************************/
/* funzioni interne alla libreria */

static Fluid_UFixPt_t UnsignedMul(Fluid_UFixPt_t A,Fluid_UFixPt_t B);
  /* Esegue la moltiplicazione tra valori unsigned in virgola fissa. */

static Fluid_FixPt_t FixedMul(Fluid_FixPt_t A,Fluid_FixPt_t B);
  /* Esegue la moltiplicazione tra valori in virgola fissa. */

static float DegToRad(float Deg);
  /* Converte gli angoli da gradi sessagesimali a radianti. */

/***************************************************************************/
/* funzioni */

static Fluid_UFixPt_t UnsignedMul(Fluid_UFixPt_t A,Fluid_UFixPt_t B)
{
  /* Esegue la moltiplicazione tra valori unsigned in virgola fissa. */
  
  
  
  Fluid_UFixPt_t	Lo,M1,M2,Hi;



  Lo=	(A & 0xffff) * (B & 0xffff);		// moltiplica le parti basse
  M1=	(A>>16) * (B & 0xffff);			// moltiplicazioni incrociate
  M2=	(A & 0xffff) * (B>>16);
  Hi=	(A>>16) * (B>>16);			// moltiplica le parti alte
  return(	(M1<<(16-RADIX_POINT)) +
  		(M2<<(16-RADIX_POINT)) +
		(Lo>>RADIX_POINT) +
		(Hi<<(32-RADIX_POINT)));
}



static Fluid_FixPt_t FixedMul(Fluid_FixPt_t A,Fluid_FixPt_t B)
{
  /* Esegue la moltiplicazione tra valori in virgola fissa. */
  
  

#if !FLUID_USE_FLOATING_POINT  
//  return((Fluid_FixPt_t)(((signed long long)A * (signed long long)B)>>16));

  if(A < 0)
  {
    if(B < 0)
      return(UnsignedMul(-A,-B));
    return(-UnsignedMul(-A,B));
  }
  if(B < 0)
    return(-UnsignedMul(A,-B));
  return(UnsignedMul(A,B));

#else
  return(A*B);
#endif
}



static float DegToRad(float Deg)
{
  /* Converte gli angoli da gradi sessagesimali a radianti. */
  
  
  
  return(Deg / 180.0 * M_PI);			// termina
}





void Fluid_Create(FluidData_t *Data,float r,float fi,Fluid_IO_t Pos)
{
  /* Crea un filtro con due poli complessi coniugati e lo resetta, i due
     poli possono essere reali coincidenti ponendo fi=0. */

  
  
  float		fGain;			// guadagno statico del filtro



  fGain=(1.0 - 2.0*r*cos(DegToRad(fi)) + r*r);
  Data->Coeff[0]=	FLOAT_TO_FIXED_POINT(fGain);
  Data->Coeff[1]=	FLOAT_TO_FIXED_POINT(2.0 * r * cos(DegToRad(fi)));
  Data->Coeff[2]=	FLOAT_TO_FIXED_POINT(-(r * r));
  Fluid_Reset(Data,Pos);			// resetta il filtro
#if ENABLE_DEBUG_MESSAGES
  printf("#%s: Coeff[0]: 0x%08lx\n",__FUNCTION__,Data->Coeff[0]);
  printf("#%s: Coeff[1]: 0x%08lx\n",__FUNCTION__,Data->Coeff[1]);
  printf("#%s: Coeff[2]: 0x%08lx\n",__FUNCTION__,Data->Coeff[2]);
#endif
  return;					// termina
}



void Fluid_Reset(FluidData_t *Data,Fluid_IO_t Pos)
{
  /* Resetta gli stati interni del filtro e termina il movimento. */
  
  
  
  Fluid_FixPt_t		Position=INT_TO_FIXED_POINT(Pos);
  
  
  
  Data->FinalPos=	Pos;			// posizione finale
  Data->Out[0]=		Position;		// resetta la storia
  Data->Out[1]=		Position;
  //Data->Out[2]=		Position;
  Data->MovCount=	0;			// movimento concluso
  return;					// termina
}



void Fluid_Start(FluidData_t *Data,Fluid_IO_t TotalSteps,Fluid_IO_t EndPos)
{
  /* Imposta la durata e la posizione finale del movimento. */
  
  
  
  Data->MovCount=	TotalSteps;		// duarata movimento
  Data->FinalPos=	EndPos;			// posizione finale
#if ENABLE_DEBUG_MESSAGES
  printf("#%s: TotalSteps: %hi  EndPos: %hi\n",__FUNCTION__,TotalSteps,EndPos);
#endif
  return;					// termina
}



int Fluid_IsCompleted(FluidData_t *Data)
{
  /* Ritorna 1 se il movimento e' completato, 0 altrimenti. */
  
  
  
  return(!Data->MovCount);			// termina
}



Fluid_IO_t Fluid_Move(FluidData_t *Data,Fluid_IO_t Input)
{
  /* Implementazione del filtro digitale. */
  
  
  
  if(Data->MovCount)				// se non zero
    Data->MovCount--;				// lo scala
  if(!Data->MovCount)				// se il movimento e' concluso
  {
    // resetta la storia
    Data->Out[1]=Data->Out[0]=	INT_TO_FIXED_POINT(Data->FinalPos);
#if ENABLE_DEBUG_MESSAGES
    printf("#%s: end reached: %hi\n",__FUNCTION__,Data->FinalPos);
#endif
    return(Data->FinalPos);			// termina con posizione finale
  }
  
  // shift della storia delle uscite
  Data->Out[2]=		Data->Out[1];
  Data->Out[1]=		Data->Out[0];

  // inserisce l'input nel filtro IIR
  Data->Out[0]=FixedMul(INT_TO_FIXED_POINT(Input),Data->Coeff[0]) +
    	       FixedMul(Data->Out[1],Data->Coeff[1]) +
    	       FixedMul(Data->Out[2],Data->Coeff[2]);
  return(FIXED_POINT_TO_INT(Data->Out[0]));	// termina
}

/* $Source$ */

