/*
    Fluid movement digital filter implementation.
    Copyright (C) 2017 2018 2021  Alessandro Accattini

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 *****************************************************************************
 *
 *
 * $Header$
 *
 *
 *
 * REVISION HISTORY:
 * $Log$
 */


/**  \file fluid.h
     \author Alessandro Accattini
     \brief Fluid movement filter implementation.
 */


#ifndef __FLUID_H
#define __FLUID_H	1


#include <stdint.h>

/***************************************************************************/
/* define pubblici */

#define FLUID_USE_FLOATING_POINT	0	// enabled if !0

/***************************************************************************/
/* macro pubbliche */

/***************************************************************************/
/* strutture pubbliche */

/** \brief Fluid filter data types.
 *
 */
#if !FLUID_USE_FLOATING_POINT
typedef int16_t		Fluid_IO_t;	///< Input/output filter data type.
typedef int32_t		Fluid_FixPt_t;	///< Fixed point value type.
typedef uint32_t	Fluid_UFixPt_t;	///< Unsigned fixed point value type.
#else
typedef int16_t		Fluid_IO_t;	///< Input/output filter data type.
typedef float		Fluid_FixPt_t;	///< Fixed point value type.
#endif


/** \brief Fluid filter data structure.
 *
 *  Fluid filter data structure is used to store filter coefficients and
 *  internal filter status.
 *  This structure must be initialized by the function \ref Fluid_Create();
 *  do not try to modify structure contents.
 */
typedef struct
{
  Fluid_FixPt_t		Out[3];		/**< Filter output history. */
  Fluid_FixPt_t		Coeff[3];	/**< Filter coefficients. */
  Fluid_IO_t		MovCount;	/**< Requested movement steps, when
  					     this counter reaches zero the
					     movement is completed. */
  Fluid_IO_t		FinalPos;	/**< Movement final position, this
  					     value is forced as output if
					     movement is completed. */
}FluidData_t;

/***************************************************************************/
/* costanti pubbliche */

extern const char	FluidVersion[];	///< Fluid library version.
extern const char	FluidBuild[];	///< Fluid library build date.

/***************************************************************************/
/* variabili pubbliche */

/***************************************************************************/
/* funzioni pubbliche */

void Fluid_Create(FluidData_t *Data,float r,float fi,Fluid_IO_t Pos);
  /**< \brief Create a new fluid data set.
       \param Data  Fluid reference data.
       \param r     Radius of complex poles in units.
       \param fi    Angle of complex poles in degrees.
       \param Pos   Filter initial position (usually 0).

       Create a new fluid data set.
       Initializes a new filter or completely reconfigures one, initial
       position is set (a call to \ref Fluid_Reset() is implied).
       Technically it creates an IIR filter with two conjugated complex poles.
       The time response is due to the poles position.
       \sa Fluid_Reset()
       \sa Fluid_Start()
   */

void Fluid_Reset(FluidData_t *Data,Fluid_IO_t Pos);
  /**< \brief Reset fluid data to given position.
       \param Data  Fluid reference data.
       \param Pos   Filter initial position.

       Reset fluid data to given position.
       Resets all the internal states of the filter as the input never moved
       from value \p Pos. The movement will result as completed if tested with
       \ref Fluid_IsCompleted() function.
       \sa Fluid_Create()
       \sa Fluid_Start()
       \sa Fluid_IsCompleted()
   */

void Fluid_Start(FluidData_t *Data,Fluid_IO_t TotalSteps,Fluid_IO_t EndPos);
  /**< \brief Starts a new movement from current position.
       \param TotalSteps  Steps expected limit fo this movement.
       \param EndPos      Desired final position.

       Starts a new movement from current position.
       This function must be called before starting a new movement.
       Movement requires some parameters as \p TotalSteps and \p EndPos:
       after \p TotalSteps steps the \p EndPos position is forced and the
       movement is completed.
       It is possible to drop the movement before reaching programmed end
       calling \ref Fluid_Reset() or simply stopping calling
       \ref Fluid_Move(). It is also possible to re-start a new movement at
       any time calling \ref Fluid_Start(), movement is restarted from current
       position.
       \sa Fluid_Create()
       \sa Fluid_Move()
   */

int Fluid_IsCompleted(FluidData_t *Data);
  /**< \brief Checks if movement is completed or not.
       \param   Data    Fluid reference data.
       \return          1 if the movement is completed, 0 otherwise.

       Checks if the running movement is completed or not, this function is
       usually called after \ref Fluid_Move() to know if the end of movement
       is reached. A completed movement cannot be continued without calling
       \ref Fluid_Start().
       \sa Fluid_Start()
       \sa Fluid_Move()
   */

Fluid_IO_t Fluid_Move(FluidData_t *Data,Fluid_IO_t Input);
  /**< \brief Moves to a new position.
       \param Data   Fluid reference data.
       \param Input  Current input, usually final position to be reached.
       \return       Filtered object position.

       Moves to a new position.
       This is the core of movement, this function is called to update the
       actual object position. Movement must be started calling 
       \ref Fluid_Start() function; a completed movement cannot be continued
       without calling \ref Fluid_Start() again.
       Once end of movement is reached, this function will return the same
       value: the end movement position programmed with \ref Fluid_Start().
       Usually this function is called with a constant \p Input value,
       the same value passed to \ref Fluid_Start() as last argument, but there
       are no restrictions on how the input value can be used. 
       \sa Fluid_Start()
       \sa Fluid_Reset()
   */

/***************************************************************************/
#endif

/* $Source$ */

