/*
    Command line magic-eye data creator.
    Copyright (C) 2017 2021  Alessandro Accattini

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 *****************************************************************************
 *
 *
 * $Header$
 *
 *
 *
 * REVISION HISTORY:
 * $Log$
 */


/***************************************************************************/
/* include vari */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>			// getopt_log()
#include <math.h>

/***************************************************************************/
/* define vari */

#define MAX_RINGS	20		// # massimo di ring per magic eye

/***************************************************************************/
/* macro varie */

/***************************************************************************/
/* strutture e tipi */

typedef unsigned long int	MemAddr_t;	// memory address (pointer)
typedef unsigned char		Vpixel_t;	// virtual pixel data type
typedef struct
{
  int		Re;			// raggio esterno
  int		Ti;			// spessore
  int		RastLines;		// raster lines
}RingDesc_t;

/***************************************************************************/
/* costanti */

static const char Help[:>=
	"Usage: %s [OPTIONS] eye_name > ../eyemap.h\n"
	"creates a map for magic-eye effect.\n\n"
	"\t-h, --help\t\tprints this help.\n"
	"\t-W, --eye-width W\tsets SDL_Surface width    (default=%i)\n"
	"\t-H, --eye-height H\tsets SDL_Surface height   (default=%i)\n"
	"\t-x, --center-x Xc\tsets center x coordinate  (default=%i)\n"
	"\t-y, --center-y Yc\tsets center y coordinate  (default=%i)\n"
	"\t-r, --radius R,T\tadds a ring of (radius, thickness)\n"
	"\t-U, --size-u U\t\tsets linear buffer width   (default=%i)\n"
	"\t-o, --ovalize O\t\tsets ovalization height (must be < H)\n"
	"\t--test-image\t\tprints out virtual image in ASCII art.\n"
	"\nIf W, H, x, y are all 0, they are automatically calculated.\n"
	"\nReport bugs to bit.trapper@gmail.com\n";

static const char		OptString[:>="+hW:H:x:y:r:U:o:";
static const struct option	OptList[:>={
	{"help",	no_argument,		NULL,'h'},
	{"eye-width",	required_argument,	NULL,'W'},
	{"eye-height",	required_argument,	NULL,'H'},
	{"center-x",	required_argument,	NULL,'x'},
	{"center-y",	required_argument,	NULL,'y'},
	{"radius",	required_argument,	NULL,'r'},
	{"size-u",	required_argument,	NULL,'U'},
	{"ovalize",	required_argument,	NULL,'o'},
	{"test-image",	no_argument,		NULL,'t'},
//	{"pippo",	required_argument,	NULL,'M'},
//	{"pappo",	optional_argument,	NULL,'M'},
	{0,0,0,0}				// end of list
};

/***************************************************************************/
/* variabili */

unsigned long int	LoadedBytes=0;		// totale bytes caricati
Vpixel_t		*DisplayMap;		// display data frame
unsigned char		*RowEn;			// array abilitazione righe

  // virtual display and surface size
int			W=0;
int			H=0;

/***************************************************************************/
/* prototipi signal handler */

/***************************************************************************/
/* prototipi funzioni */

static void RowsEnable(unsigned char *Enable,int Dx,int Dy);
  /* Seleziona le righe visibili. */

static int MaxRadius(const RingDesc_t *Ring,int Rings);
  /* Determina il raggio max tra i ring che sono stati dichiarati. */

static void DisplayFill(Vpixel_t Color);
  /* Riempie il display virtuale col colore indicato. */

static void DisplayPutPixel(int x,int y,Vpixel_t Color);
  /* Colora il pixel indicato col colore specificato. */

static Vpixel_t DisplayGetPixel(int x,int y);
  /* Restituisce il colore del pixel indicato. */

static void FloodFill(int x,int y,Vpixel_t OldColor,Vpixel_t NewColor);
  /* Esegue l'algoritmo ricorsivo Flood-fill alla posizione indicata. */

static void DisplayPaint(int x,int y,Vpixel_t Color);
  /* Colora col colore indicato a partire dal punto specificato. */

static void DisplayCircle(int x,int y,int r,Vpixel_t Color);
  /* Disegna un cerchio vuoto alle coordinate indicate. */

static int CreateRing(int Cx,int Cy,int Re,int Ti,int U,int RinfId,
		      const char *Name);
  /* Crea un ring e stampa da definizione sullo stdout, ritorna il numero di
     rester lines che compongono il ring. */

/***************************************************************************/
/* funzioni */

static void RowsEnable(unsigned char *Enable,int Dx,int Dy)
{
  /* Seleziona le righe visibili. */


  /*
     Utilizza l'algoritmo di Bresenham per ottenre una compressione
     (stretching) delle righe originali dell'immagine virtuale.
     Dx rappresenta il numero originale + 1 di righe nell'immagine,
     Dy rappresenta il numero compresso + 1 di righe nell'immagine.

     Ad ogni iterazione dell'algoritmo se si ha un salto verticale
     allora la riga dell'immegine viene preservata, altrimenti viene
     disabilitata.
  */


  int	d,k1,k2;



  d=(Dy * 2) - Dx;				// costanti di Bresenham
  k1=Dy * 2;
  k2=2 * (Dy - Dx);

  while(--Dx)
  {
    if(d < 0)
    {
      // Bresenham non ha un salto, disabilita la riga
      d+=k1;
      *Enable++=0;				// disabilita la riga
    }
    else
    {
      // Bresenham ha un salto, abilita la riga
      d+=k2;
      *Enable++=1;				// abilita la riga
    }
  }
  return;					// termina
}



static int MaxRadius(const RingDesc_t *Ring,int Rings)
{
  /* Determina il raggio max tra i ring che sono stati dichiarati. */



  int		Max=0;



  while(Rings--)
  {
    if(Ring->Re > Max)				// se e' piu' grande
      Max=Ring->Re;				// cambia il max
    Ring++;					// next ring
  }
  return(Max);					// termina
}



static void DisplayFill(Vpixel_t Color)
{
  /* Riempie il display virtuale col colore indicato. */
  
  
  
  Vpixel_t	*Ptr=DisplayMap;
  int		i=W * H;
  
  
  
  while(i--)
    *Ptr++=Color;
  return;					// termina
}



static void DisplayPutPixel(int x,int y,Vpixel_t Color)
{
  /* Colora il pixel indicato col colore specificato. */
  
  
  
  if((x < 0) || (x >= W))			// se fuori dal display
    return;					// termina immediatamente
  if((y < 0) || (y >= H))			// se fuori dal display
    return;					// termina immediatamente
  DisplayMap[(y * W) + x]=Color;		// colora un pixel
  return;					// termina
}



static Vpixel_t DisplayGetPixel(int x,int y)
{
  /* Restituisce il colore del pixel indicato. */



  if((x < 0) || (x >= W))			// se fuori dal display
    return 0;					// fake value
  if((y < 0) || (y >= H))			// se fuori dal display
    return 0;					// fake value
  return(DisplayMap[(y * W) + x]);		// termina
}



static void FloodFill(int x,int y,Vpixel_t OldColor,Vpixel_t NewColor)
{
  /* Esegue l'algoritmo ricorsivo Flood-fill alla posizione indicata. */
  
  
  
  if((x < 0) || (x >= W))			// se fuori dal display
    return;					// termina
  if((y < 0) || (y >= H))			// se fuori dal display
    return;					// termina
  if(DisplayGetPixel(x,y) != OldColor)		// se abbiamo sconfinato
    return;					// termina
  if(DisplayGetPixel(x,y) == NewColor)		// se e' gia' colorato
    return;					// termina

  DisplayPutPixel(x,y,NewColor);		// colora il pixel attuale
  FloodFill(x,y-1,OldColor,NewColor);		// pixel a nord
  FloodFill(x+1,y,OldColor,NewColor);		// pixel a est
  FloodFill(x,y+1,OldColor,NewColor);		// pixel a sud
  FloodFill(x-1,y,OldColor,NewColor);		// pixel a ovest
  return;					// termina
}



static void DisplayPaint(int x,int y,Vpixel_t Color)
{
  /* Colora col colore indicato a partire dal punto specificato. */
  
  
  
  Vpixel_t	OldColor;
  
  
  
  OldColor=DisplayGetPixel(x,y);		// colore originale nell'area
  FloodFill(x,y,OldColor,Color);		// avvia la colorazione
  return;					// termina
}



static void DisplayCircle(int Cx,int Cy,int r,Vpixel_t Color)
{
  /* Disegna un cerchio vuoto alle coordinate indicate. */


  /*
     Radius = 1 circle

      #
     # #
      #
  */


  int	d,x,y;
  
  
  
  // inizializza il cerchio
  d=3 - (2 * r);
  x=0;
  y=r;
  
  // traccia il cerchio
  do
  {
    DisplayPutPixel(Cx + x,Cy + y,Color);
    DisplayPutPixel(Cx + x,Cy - y,Color);
    DisplayPutPixel(Cx - x,Cy + y,Color);
    DisplayPutPixel(Cx - x,Cy - y,Color);
    DisplayPutPixel(Cx + y,Cy + x,Color);
    DisplayPutPixel(Cx + y,Cy - x,Color);
    DisplayPutPixel(Cx - y,Cy + x,Color);
    DisplayPutPixel(Cx - y,Cy - x,Color);
    
    if(d < 0)
      d+=(4*x)+6;
    else
    {
      d+=(x-y)*4+10;
      y--;
    }
    x++;
  }while(x <= y);
  return;					// termina
}



static int Mapping(int Cx,int Cy,int x,int y,int X)
{
  /* Rimappa il pixel indicato all'interno del vettore lineare. */
  
  
  
  /*
     Remapping dalla corona circolare al rettangolo:
      x=fi * X / (2pi)
      y=(r-Re) * Y / (Ri-Re)
  
     Questa regola viene applicata in modo semplificato essendo il rettangolo
     monodimensionale (y non viene calcolato).
  */
  
  
  
  double	fi;			// argomento della coordinata polare
  
  
  
  // coordinate cartesiane all'interno del cerchio
//  printf("%s: Cx=%i, Cy=%i, x=%i, y=%i, X=%i\n",__FUNCTION__,Cx,Cy,x,y,X);
  x-=Cx;
  y-=Cy;
  y=-y;					// inverte l'asse y
  
  // determina le coordinate polari
  fi=atan((double)x/(double)y);		// arco-cotangente
//  printf("%s: x=%i, y=%i, fi=%6.3f\n",__FUNCTION__,x,y,fi);
//  if(fi < 0.0)
//    fi+=M_PI+M_PI;			// fi compreso tra 0 e 2 pi
  if(y < 0.0)
    fi+=M_PI;
  else if(x < 0.0)
    fi+=M_PI+M_PI;
//  printf("%s: x=%i, y=%i, fi=%6.3f\n",__FUNCTION__,x,y,fi);
  
  // proietta dal cerchio al rettangolo
  return((int)(fi * (double)X / (2.0 * M_PI)));
}



static int CreateRing(int Cx,int Cy,int Re,int Ti,int U,int RingId,
		      const char *Name)
{
  /* Crea un ring e stampa da definizione sullo stdout, ritorna il numero di
     rester lines che compongono il ring. */


  /*
     La corona circolare viene ottenuta riempiendo tutto il display
     col colore 1, vengono tracciati i due cerchi con colore 2 e poi
     l'esterno della corona viene svuotato con colore 0.
  */

  
  int		Ri;				// raggio interno
  int		i,j,k,Row;
  int		SegX1=0;
  int		SegY1=0;
  int		SegX2=0;
  int		RastLines=0;
  Vpixel_t	Color;				// pixel color
  unsigned char	Inside=0;			// flag inside/outside
  unsigned char	ColorFound;			// flag color found on row



  // crea il ring di riferimento
  Ri=Re - Ti + 1;				// raggio interno
  DisplayFill(1);				// svuota tutto il display
  DisplayCircle(Cx,Cy,Re,2);			// raggio esterno
  DisplayCircle(Cx,Cy,Ri,2);			// raggio interno
  DisplayPaint(0  ,0  ,0);			// esterno alto sinistro
  DisplayPaint(W-1,0  ,0);			// esterno alto destro
  DisplayPaint(0  ,H-1,0);			// esterno basso sinistro
  DisplayPaint(W-1,H-1,0);			// esterno basso destro
  DisplayPaint(Cx,Cy,0);			// interno

  // scanning orizzontale dell'immagine per la generazione del remap dei pixel
  for(j=0;j<H;j++)
  {
    if(!RowEn[j])			// se la riga e' disabilitata
      continue;				// ignora la riga
    for(i=0;i<W;i++)
    {
      Color=DisplayGetPixel(i,j);	// colore del pixel attuale
      if(!Inside && Color)		// se e' il primo pieno che incontri
      {
	// inizio segmento
	Inside=1;			// ora siamo dentro
	SegX1=i;			// salva le coordinate del pixel
	SegY1=j;
	printf("static const Index_t\tMap_%s_%02i_%04i[:>={",
		Name,RingId,++RastLines);
	continue;
      }
      if(Inside && !Color)		// se e' il primo vuoto che incontri
      {
	// fine segmento
	Inside=0;			// ora siamo fuori
	SegX2=i-1;			// salva le coordinate del pixel
//	printf("// SegX1=%i, SegY1=%i, SegX2=%i\n",SegX1,SegY1,SegX2);
	for(k=SegX1;k<=SegX2;k++)
	  printf("%i,",Mapping(Cx,Cy,k,SegY1,U));
	printf("};\n");
	continue;
      }
    }
  }
  
  // scanning orizzontale dell'immagine per la generazione del raster
  RastLines=0;				// azzera il contatore
  Row=0;
  printf("\n\n\nstatic const Rast_t\tRaster_%s_%02i[:>={\n",Name,RingId);
  for(j=0;j<H;j++)
  {
    if(!RowEn[j])			// se la riga e' disabilitata
      continue;				// ignora la riga
    ColorFound=0;			// no color found
    for(i=0;i<W;i++)
    {
      Color=DisplayGetPixel(i,j);	// colore del pixel attuale
      if(!Inside && Color)		// se e' il primo pieno che incontri
      {
	Inside=1;			// ora siamo dentro
	ColorFound=1;			// found a valid color
	SegX1=i;			// salva le coordinate del pixel
	SegY1=Row; //j;
	continue;
      }
      if(Inside && !Color)		// se e' il primo vuoto che incontri
      {
	Inside=0;			// ora siamo fuori
	SegX2=i-1;			// salva le coordinate del pixel
	printf("\t\t\t{%3i,%3i,%3i,Map_%s_%02i_%04i},\n",
	       SegX1,SegY1,SegX2,Name,RingId,++RastLines);
	continue;
      }
//      if(!ColorFound)			// if row is empty
//        Row++;				// increase the counter
    }
    Row++;				// riga usata
  }
  printf("};\n\n\n");
  return(RastLines);			// termina
}





/***************************************************************************/

int main(int argc,char **argv)
{
  RingDesc_t	Ring[MAX_RINGS];		// descrittore dei ring
  int		Rings=0;			// # di ring dichiarati
  int		Option;				// line command option
  int		OptIdx;				// option index
  int		Cx=0;				// centro dell'anello
  int		Cy=0;
  int		Ovalization=0;			// # righe in uscita
  int		U=200;				// dimensione array lineare
  int		i,j;
  char		*EyeName;			// nome del magic eye
  unsigned char	TestImage=0;			// test image flag



  // legge le opzioni presenti sulla riga di comando
  while((Option=getopt_long(argc,argv,OptString,OptList,&OptIdx)) >= 0)
  {
    switch(Option)
    {
      case 'h':
        printf(Help,argv[0],W,H,Cx,Cy,U);
	return 0;			// termina
      case 'W':
        sscanf(optarg,"%i",&W);		// width
	break;
      case 'H':
        sscanf(optarg,"%i",&H);		// height
	break;
      case 'x':
        sscanf(optarg,"%i",&Cx);	// coordinata X centro del ring
	break;
      case 'y':
        sscanf(optarg,"%i",&Cy);	// coordinata Y centro del ring
	break;
      case 'o':
        sscanf(optarg,"%i",&Ovalization);	// abilita l'ovalizzazione
	break;
      case 't':
        TestImage=1;
	break;
      case 'r':
        if(Rings >= MAX_RINGS)
	{
	  fputs("Maximal rings count reached.\n",stderr);
	  return 1;			// termina
	}

	// aggiunge un nuovo ring al magic eye
        if(sscanf(optarg,"%i,%i",&Ring[Rings].Re,&Ring[Rings].Ti) != 2)
	{
	  fputs("Error while deconding radius option argument.\n",stderr);
	  return 1;			// termina
	}
	Rings++;			// conta il nuovo ring
	break;
      case 'U':
        sscanf(optarg,"%i",&U);		// limite array lineare
	break;
      default:
        fputs("Unknown option, use -h to see help.\n",stderr);
	return 2;			// termina
    }
  }
  
  // verifica che sia stato dichiarato almeno un ring
  if(!Rings)					// se non ce ne sono
  {
    fputs("No rings have been declared.\n",stderr);
    return 5;					// termina
  }

  // determina in modo automatico la dim. della superficie ed il centro
  if(!W && !H && !Cx && !Cy)			// se sono tutti nulli
  {
    int		RMax;



    RMax=MaxRadius(Ring,Rings);			// trova il raggio max
    Cx=RMax;					// determina il centro
    Cy=RMax;
    W=(RMax + 1) * 2;				// determina la superficie
    H=(RMax + 1) * 2;
    printf("//\n// Automatically calculated surface size: W=%i, H=%i\n",W,H);
    printf("// Automatically calculated eye center: Cx=%i, Cy=%i\n//\n",Cx,Cy);
  }
  
//  // processa tutti gli altri parametri passati
//  for(;optind<argc;++optind)			// li passa tutti
//    VoiceDir=argv[optind];			// tiene l'ultimo

  // cerca il nome del magic eye dopo tutti i parametri
  if(optind >= argc)				// se manca il nome
  {
    fputs("Magic eye name not found.\n",stderr);
    return 3;					// termina con disastro
  }
  EyeName=argv[optind];				// punta al nome dell'eye
  
  // alloca un buffer grande come tutto il display, 1 byte / pixel
  if((DisplayMap=malloc((W * H) + H)) == NULL)
  {
    fputs("Unable to allocate image buffer.\n",stderr);
    return 4;					// termina con disastro
  }
  RowEn=&DisplayMap[W * H];			// array abilitazione righe

  // abilita le righe in funzione dell'ovalizzazione
  if(!Ovalization || (Ovalization > H))		// se disabilitata
    Ovalization=H;				// azzera l'ovalizzazione
  RowsEnable(RowEn,H+1,Ovalization+1);		// seleziona le righe visibili

  // intesta l'output
  printf("//\n// This code is automatically generated by %s.c\n//\n",argv[0]);
  printf("// Declared rings:\n");
  for(i=0;i<Rings;i++)
    printf("// ring %i: Re=%i,Th=%i\n",i+1,Ring[i].Re,Ring[i].Ti);
  printf("//\n//\n\n");

  // processa tutti i ring dichiarati
  for(i=0;i<Rings;i++)
  {
    Ring[i].RastLines=CreateRing(Cx,Cy,		// centro dell'anello
    				 Ring[i].Re,	// raggio esterno
				 Ring[i].Ti,	// spessore
				 U,		// dimensione lineare
				 i+1,		// ring ID
				 EyeName);	// eye unique name
  }

  // crea l'array lineare di pixel
  printf("static Color_t\t\t\t%s_Pixels[%i];\n\n\n",EyeName,U*Rings);

  // crea l'array di SDL_Surface
  printf("static SDL_Surface\t\t*%s_Img[2]={NULL,NULL};\n\n\n",EyeName);

  // crea la tabella dei raster
  printf("static const RasterDesc_t\t%s_RasterList[:>={\n",EyeName);
  for(i=0;i<Rings;i++)
    printf("\t\t\t\t{.Lines=%i,.Raster=Raster_%s_%02i},\n",
    	   Ring[i].RastLines,EyeName,i+1);
  printf("};\n\n\n");

  // crea il magic eye descriptor
  printf("const MagicEye_t\t\t%s={\n",EyeName);
  printf("\t\t\t\t.w=%i,\n",W);
  printf("\t\t\t\t.h=%i,\n",Ovalization);
  printf("\t\t\t\t.u=%i,\n",U);
  printf("\t\t\t\t.v=%i,\n",Rings);
  printf("\t\t\t\t.Raster=%s_RasterList,\n",EyeName);
  printf("\t\t\t\t.Pixel=%s_Pixels,\n",EyeName);
  printf("\t\t\t\t.Imgs=%s_Img,\n",EyeName);
  printf("};\n\n\n");
  
  // test dell'immagine
  if(TestImage)
  {
    for(j=0;j<H;j++)
    {
      if(!RowEn[j])				// se la riga e' disabilitata
	continue;				// ignora la riga
      printf("%2i  ",j);
      for(i=0;i<W;i++)
      {
        if((j == Cy) && (i == Cx))
	{
	  putchar('+');				// eye center
	  continue;
	}
	if(DisplayGetPixel(i,j) == 1)
	{
	  putchar('O');				// filler
	  continue;
	}
	if(DisplayGetPixel(i,j) == 2)
	{
	  putchar('#');				// circle border
	  continue;
	}
	putchar('.');				// empty space
      }
      putchar('\n');
    }
  }
  
  free(DisplayMap);				// libera il buffer
  return 0;					// termina
}

/* $Source$ */
 
