/*
    Magic eye handling functions.
    Copyright (C) 2017 2018 2019 2021  Alessandro Accattini

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 *****************************************************************************
 *
 *
 * $Header$
 *
 *
 *
 * REVISION HISTORY:
 * $Log$
 */


/***************************************************************************/
/* include vari */

#include "magiceye.h"				// functions prototypes

/***************************************************************************/
/* define interni alla libreria */

/***************************************************************************/
/* strutture contenute all'interno della libreria */

/***************************************************************************/
/* private magic eye data struct declaration */

#include "eyemap.c"			// eye mapping

/***************************************************************************/
/* variabili interne alla libreria */

/***************************************************************************/
/* funzioni interne alla libreria */

static int Mod(int Value,int Module);
  /* Ritorna Value limitato al modulo indicato. */

static Color_t *PixelPtr(SDL_Surface *Img,Coord_t x,Coord_t y);
  /* Ritorna un puntatore al pixel indicato all'interno della surface. */

static void Perspection(const Rast_t *LinePtr,int Lines,Color_t *Linear,
			int Rotation,int Usize,SDL_Surface *Img);
  /* Realizza il rendering di un singolo anello del magic eye. */

/***************************************************************************/
/* funzioni */

static int Mod(int Value,int Module)
{
  /* Ritorna Value limitato al modulo indicato. */



  while(Value < 0)			// se e' negativo
    Value+=Module;			// lo riporta in carreggiata
  while(Value >= Module)		// se supera il primo giro
    Value-=Module;			// lo riporta in carreggiata
  return(Value);			// termina
}



static Color_t *PixelPtr(SDL_Surface *Img,Coord_t x,Coord_t y)
{
  /* Ritorna un puntatore al pixel indicato all'interno della surface. */



  Color_t	*Line;			// inizio linea data dalla Y
  unsigned char	*Byte;			// puntatore a byte



  Byte=(unsigned char*)Img->pixels;	// punta all'inizio del frame
  Line=(Color_t*)&Byte[y * Img->pitch];	// punta all'inizio della riga
  return(&Line[x]);			// termina
}



static void Perspection(const Rast_t *LinePtr,int Lines,Color_t *Linear,
			int Rotation,int Usize,SDL_Surface *Img)
{
  /* Realizza il rendering di un singolo anello del magic eye. */



  const Index_t	*MapPtr;			// remapping nella linea
  Index_t	Rotated;			// index after rotation
  int		Pixels;				// totale di pixel nella riga
  Color_t	*DrawPtr;			// puntatore al pixel scritto



  do						// per tutte le righe
  {
    Pixels=LinePtr->x2 - LinePtr->x1 + 1;	// # pixel nella riga
    MapPtr=LinePtr->ReMap;			// remapping della riga
    
    // punta al primo pixel della riga
    if((DrawPtr=PixelPtr(Img,LinePtr->x1,LinePtr->y1)) != NULL)
    {
      // riempie una riga del magic eye
      do					// per tutti i pixel
      {
        Rotated=Mod(*MapPtr++ - Rotation,Usize);// corregge l'index ruotandolo
	*DrawPtr++=Linear[Rotated];		// colora la surface
      }while(--Pixels);
    }
    LinePtr++;					// next raster line
  }while(--Lines);
  return;					// termina
}





SDL_Surface *MagicEye_CreateSurfaces(const MagicEye_t *EyeDesc)
{
  /* Creates the required magic eye surfaces. */



  Color_t		*Pix;		// pixel pointer
  int			i;



  // prima di tutto azzeriamo le immagini create
  EyeDesc->Imgs[0]=NULL;
  EyeDesc->Imgs[1]=NULL;

  // crea la surface per il magic eye
  if((EyeDesc->Imgs[0]=SDL_CreateRGBSurface(SDL_SWSURFACE | SDL_ASYNCBLIT,
  				EyeDesc->w,
  				EyeDesc->h,
				16,			// bit/pixel
				0xf800,			// R mask
				0x07e0,			// G mask
				0x001f,			// B mask
				0x0000)) == NULL)	// A mask
    return(NULL);			// termina con fallimento
  
  // colora la surface di nero
  i=EyeDesc->Imgs[0]->pitch / sizeof(Color_t) * EyeDesc->Imgs[0]->h;
  Pix=EyeDesc->Imgs[0]->pixels;		// punta ai pixel
  do
  {
    *Pix++=0;				// riempie di nero
  }while(--i);

  // crea la surface per l'array rettangolare di pixel
  if((EyeDesc->Imgs[1]=SDL_CreateRGBSurfaceFrom(EyeDesc->Pixel,	// pixels array
  				EyeDesc->u,		// width
				EyeDesc->v,		// height
				16,			// bit/pixel
  				EyeDesc->u * sizeof(Color_t),	// byte/row
				0xf800,			// R mask
				0x07e0,			// G mask
				0x001f,			// B mask
				0x0000)) == NULL)	// A mask
  {
    SDL_FreeSurface(EyeDesc->Imgs[0]);	// distrugge la surface
    return(NULL);			// termina con fallimento
  }

  // riempie l'array di nero  
  MagicEye_Fill(EyeDesc,0x0000);
  return(EyeDesc->Imgs[0]);		// termina con successo
}



SDL_Surface *MagicEye_GetPixmap(const MagicEye_t *EyeDesc)
{
  /* Lets you access pixels map as an SDL_Surface. */



  return(EyeDesc->Imgs[1]);		// pixels map as SDL_Surface
}



void MagicEye_DestroySurfaces(const MagicEye_t *EyeDesc)
{
  /* Destroies magic-eye surfaces. */



  if(EyeDesc->Imgs[0] != NULL)		// se esiste
    SDL_FreeSurface(EyeDesc->Imgs[0]);	// distrugge la surface
  if(EyeDesc->Imgs[1] != NULL)		// se esiste
    SDL_FreeSurface(EyeDesc->Imgs[1]);	// distrugge la surface
  EyeDesc->Imgs[0]=NULL;		// adesso di sicuro non ci sono piu'
  EyeDesc->Imgs[1]=NULL;
  return;				// termina
}



void MagicEye_PutPixel(const MagicEye_t *EyeDesc,int u,int v,Color_t Color)
{
  /**< \brief  place a pixel in the rectangular array.
       \param EyeDesc  the magic eye descriptor.
       \param u  the horizontal coordinate in the rectangular array, this
       		 coordinate is equivalent to the agular coordinate in the eye.
       \param v  the vertical coordinate in the rectangular array, this
       		 coordinate is oriented radially in the eye, starting from the
		 external ring toward to the centre of the eye.
       \param Color  the pixel color.

       This function simply place a pixel in the rectangular array, no
       changes to the magic eye are implied, to update the SDL_Surface you
       need to explicitely call MagicEye_Perspection().
   */



  Color_t	*Linear=EyeDesc->Pixel;		// punta all'array lineare



  // verifica che le coordinate siano all'interno del vettore
  if(u < 0)
    return;					// fermati qui
  if(v < 0)
    return;					// fermati qui
  if(u >= EyeDesc->u)
    return;					// fermati qui
  if(v >= EyeDesc->v)
    return;					// fermati qui
  
  Linear[(v * EyeDesc->u) + u]=Color;		// colora un pixel
  return;					// termina
}



Color_t MagicEye_GetPixel(const MagicEye_t *EyeDesc,int u,int v)
{
  /**< \brief  reads back the pixel color.
       \param EyeDesc  the magic eye descriptor.
       \param u  the horizontal coordinate in the rectangular array, this
       		 coordinate is equivalent to the agular coordinate in the eye.
       \param v  the vertical coordinate in the rectangular array, this
       		 coordinate is oriented radially in the eye, starting from the
		 external ring toward to the centre of the eye.
       \return  the pixel color.

       This function can be used to read back a pixel color (see
       MagicEye_PutPixel()).
   */
  
  
  
  Color_t	*Linear=EyeDesc->Pixel;		// punta all'array lineare



  // verifica che le coordinate siano all'interno del vettore
  if(u < 0)
    return 0;					// fermati qui
  if(v < 0)
    return 0;					// fermati qui
  if(u >= EyeDesc->u)
    return 0;					// fermati qui
  if(v >= EyeDesc->v)
    return 0;					// fermati qui
  
  return(Linear[(v * EyeDesc->u) + u]);		// restituisce il colore
}



void MagicEye_Fill(const MagicEye_t *EyeDesc,Color_t Color)
{
  /**< \brief  fills the whole rectangular array with the same color.
       \param EyeDesc  the magic eye descriptor.
       \param Color  the color to fill with.

       Fills the whole rectangular array at once, usefull for example to
       clear all pixel in a single call.
   */
  
  
  
  Color_t	*Linear=EyeDesc->Pixel;		// punta all'array lineare
  int		i=EyeDesc->u * EyeDesc->v;	// totale pixel



  while(i--)					// fin che ce n'e'
    *Linear++=Color;				// colora
  return;					// termina
}



void MagicEye_Update(const MagicEye_t *EyeDesc,int Rotate[])
{
  /**< \brief  updates the magic eye surface.
       \param EyeDesc  the magic eye descriptor.
       \param Rotate  a rotation steps array.
       \param Img  the SDL_Surface in with the magic eye is drawn.

       This function creates the real magic eye: after all rectangular array
       draw you need to call this function to update the SDL_Surface.
   */



  int		i,Rotation;



  for(i=0;i<EyeDesc->v;i++)
  {
    if(Rotate == NULL)				// se non eiste l'array
      Rotation=0;				// nessuna rotazione
    else
      Rotation=Rotate[i];			// preleva la rot. del ring
    Perspection(EyeDesc->Raster[i].Raster,	// raster table
    		EyeDesc->Raster[i].Lines,	// # raster lines
		&EyeDesc->Pixel[i*EyeDesc->u],	// linear pixel vector
		Rotation,			// rotazione anello
		EyeDesc->u,			// dimensione asse U
		EyeDesc->Imgs[0]);		// surface
  }
  return;					// termina
}

/* $Source$ */

