/*
    Magic eye handling functions.
    Copyright (C) 2017 2018 2019 2021  Alessandro Accattini

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 *****************************************************************************
 *
 *
 * $Header$
 *
 *
 *
 * REVISION HISTORY:
 * $Log$
 */

/**  \file magiceye.h
 *   \author Alessandro Accattini
 *   \date 24 Apr 2017
 *   \date 21 Nov 2019
 *   \brief Magic eye handling functions header file.
 *
 *   In mathematical language a magic eye is a circular projection of a
 *   rectangular pixels array, setting the color of a pixel in the pixels
 *   array is equivalent to change the color of a little circular sector of
 *   the magic eye.
 *
 *   In common language a magic eye is a set of concentric rings of different
 *   thickness and shape (each ring can be coloured or incomplete) rotating
 *   in different ways with respect to each other.
 *
 *   This file provides all functions to create, move, modify and animate one 
 *   or more magic eyes.
 *   Each magic eye is based on a pre-built data set that is considered to be
 *   constant, this means the rings of a magic eye can rotate, can disappear,
 *   can be reshaped, but rings thickness and diameter cannot be changed.
 */


#ifndef __MAGICEYE_H
#define __MAGICEYE_H	1


#include <stdint.h>
#include <SDL/SDL.h>				// SDL library data types

/***************************************************************************/
/* define pubblici */

/***************************************************************************/
/* macro pubbliche */

/***************************************************************************/
/* strutture pubbliche */

/** \brief Pixel color data type.
 *
 */
typedef uint16_t	Color_t;


/** \brief Array index data type.
 *
 */
typedef uint32_t	Index_t;


/** \brief Plane coordinate data type.
 *
 */
typedef int16_t		Coord_t;


/** \brief Single rast data type.
 *
 */
typedef struct
{
  Coord_t		x1;		///< Line horizontal begin coordinate.
  Coord_t		y1;		///< Line vertical begin coordinate.
  Coord_t		x2;		///< Line horizontal end coordinate.
  const Index_t		*ReMap;		///< Line remapping.
}Rast_t;


/** \brief Raster descriptor structure.
 *
 * Raster descriptor defines a full ring inside the magic eye, it is made up
 * of lines of scan that all together draw the ring shape.
 */
typedef struct
{
  int			Lines;		///< Number of raster lines.
  const Rast_t		*Raster;	///< The raster lines table.
}RasterDesc_t;


/** \brief Magic eye descriptor structure.
 *
 * Magic eye descriptor is the main data structure containing all
 * infrormations about the magic eye hitself. Magic eye is not run
 * time modifiable, so this one and all descendant structures are
 * constants.
 * Single magic eye is composed of more concentric rings, each ring is
 * represented by a raster descriptor of type \ref RasterDesc_t.
 * \ref Raster is a pointer to an array of raster descriptors, we have as
 * many descriptors (aka rings) as \ref v dimension.
 */
typedef struct
{
  Uint16		w;	/**< SDL_Surface width. */
  Uint16		h;	/**< SDL_Surface height. */
  Uint16		u;	/**< Pixels map width, number of sectors. */
  Uint16		v;	/**< Pixels map height, number of rings. */
  const RasterDesc_t	*Raster;/**< The raster descriptors table, this table
  				     has many entries as v size indicates. */
  Color_t		*Pixel;	/**< Pixel linear array pointer. */
  SDL_Surface		**Imgs; /**< Array of SDL_Surfaces pointer. */
}MagicEye_t;

/***************************************************************************/
/* variabili pubbliche */

/***************************************************************************/
/* funzioni pubbliche */

SDL_Surface *MagicEye_CreateSurfaces(const MagicEye_t *EyeDesc);
  /**< \brief  Creates the required magic-eye surfaces.
       \param EyeDesc    The magic eye descriptor.
       \return           On success, returns the magic-eye main surface.

       Creates the required magic eye surfaces.
       You need to create an SDL_Surface of suitable size in order to render
       the magic-eye in it, this function do the job.
       The whole magic-eye descriptor is allocated as const, the variable
       parts needs to be allocated run-time.
       The main surface, used while rendering, is returned by this function;
       the other surface mey be accessed calling \ref MagicEye_GetPixmap().
       \sa MagicEye_Update()
       \sa MagicEye_GetPixmap()
       \sa MagicEye_DestroySurfaces()
   */

SDL_Surface *MagicEye_GetPixmap(const MagicEye_t *EyeDesc);
  /**< \brief  Lets you access pixels map as an SDL_Surface.
       \param EyeDesc    The magic eye descriptor.
       \return           On success, returns the pixels map surface.
  
       Lets you access pixels map as an SDL_Surface.
       You may access the pixels map calling \ref MagicEye_PutPixel() and
       \ref MagicEye_GetPixel(), or you may access it as an SDL_Surface; this
       second way lets you load images in it using SDL APIs, or copy/move some
       parts using SDL_BlitSurface().
   */

void MagicEye_DestroySurfaces(const MagicEye_t *EyeDesc);
  /**< \brief  Destroies magic-eye surfaces.
       \param EyeDesc    The magic eye descriptor.

       Destroies magic-eye surfaces.
       Surfaces automatically allocated by \ref MagicEye_CreateSurfaces()
       needs to be freed before leaving.
       \sa MagicEye_CreateSurfaces()
   */

void MagicEye_PutPixel(const MagicEye_t *EyeDesc,int u,int v,Color_t Color);
  /**< \brief  Places a pixel in the pixel map.
       \param EyeDesc    The magic eye descriptor.
       \param u          The horizontal coordinate in the pixel map.
                         It is equivalent to the agular coordinate.
       \param v          The vertical coordinate in the pixel map.
                         It is equivalent to choose a ring in the eye starting
			 from the most external ring.
       \param Color      The pixel color.

       Places a pixel in the pixel map.
       It simplyfies placing a pixel in the pixel map, to update changes in
       the magic eye \ref MagicEye_Update() must be called.
       \sa MagicEye_Fill()
       \sa MagicEye_GetPixel()
   */

Color_t MagicEye_GetPixel(const MagicEye_t *EyeDesc,int u,int v);
  /**< \brief  Reads back the pixel color.
       \param EyeDesc    The magic eye descriptor.
       \param u          The horizontal coordinate in the pixel map.
                         It is equivalent to the agular coordinate.
       \param v          The vertical coordinate in the pixel map.
                         It is equivalent to choose a ring in the eye starting
			 from the most external ring.
       \return           The pixel color.

       Reads back the pixel color.
       \sa MagicEye_PutPixel()
       \sa MagicEye_Fill()
   */

void MagicEye_Fill(const MagicEye_t *EyeDesc,Color_t Color);
  /**< \brief  Fills the whole pixel map with the same color.
       \param EyeDesc    The magic eye descriptor.
       \param Color      The color to fill with.

       Fills the whole pixel map with the same color.
       Usefull for example to clear the pixel map on start-up.
       \sa MagicEye_PutPixel()
       \sa MagicEye_GetPixel()
   */

void MagicEye_Update(const MagicEye_t *EyeDesc,int Rotate[]);
  /**< \brief  Updates the magic eye on SDL_Surface.
       \param EyeDesc    The magic eye descriptor.
       \param Rotate     Rings rotation array (can be NULL).

       Updates the magic eye on SDL_Surface.
       All pixels in the pixels map are transformed in a piece of magic eye
       and the SDL_Surface is updated.
       While updating magic eye all its rings can be rotated, \p Rotate is
       used to acieve this: it is an array of int, one for each eye ring.
       Ring rotation is measured in sectors: if the magic eye has 128 sectors
       a rotation of 64 means half turn in clockwise direction.
       \p Rotate can be used to make eye animation whithout changing pixels
       map. \p Rotate can be NULL if unused.
   */

/***************************************************************************/
#endif

/* $Source$ */

