/*
    Sequencer implementation.
    Copyright (C) 2021  Alessandro Accattini

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 *****************************************************************************
 *
 *
 * $Header$
 *
 *
 *
 * REVISION HISTORY:
 * $Log$
 */

/**  \file sequencer.h
 *   \author Alessandro Accattini
 *   \brief Sequencer implementation and control APIs.
 */


#ifndef __SEQUENCER_H
#define __SEQUENCER_H	1


#include <stdint.h>

/***************************************************************************/
/* define pubblici */

/***************************************************************************/
/* macro pubbliche */

/** \brief Sequencer creation macro.
 *
 *  Sequencer is statically allocated and initialized using this macro.
 *  It is not a good idea to modify directly sequencer data, please use
 *  dedicated APIs.
 */
#define CREATE_SEQUENCER(SeqName,Sequence)	Sequencer_t SeqName= \
							{Sequence,0,0}

/***************************************************************************/
/* strutture pubbliche */

/** \brief Sequence cut data type.
 *
 *  A sequence is a set of cuts; each cut is a function that is called until
 *  it returns a zero value. When a sequence cut returns a zero value the cut
 *  itself is never called, the subsequent cut is called instead.
 *  If the end of sequence is reached, it is restarted from the first cut.
 *  A cut function is called with two arguments: a user definable data pointer
 *  and an incremental integer, this value is always 0 when the cut function
 *  is called the first time.
 */
typedef unsigned int (*SeqCut_t)(void*,unsigned int);



/** \brief Sequencer data type.
 *
 */
typedef struct
{
  const SeqCut_t	*Sequence;	///< NULL terminated sequence.
  unsigned int		Active;		///< Current active sequence cut.
  unsigned int		Frame;		///< Current cut frame counter.
}Sequencer_t;

/***************************************************************************/
/* variabili pubbliche */

/***************************************************************************/
/* funzioni pubbliche */

void Sequencer_Run(Sequencer_t *Sequencer,void *Data);
  /**< \brief Runs sequencer's sequence single cut step.
       \param Sequencer    Sequencer to be ativated.
       \param Data         User definable data.

       Runs sequencer's sequence single cut step.
       It executes a single step in a sequencer sequence; if end of sequence
       is reached, the sequence is automatically re-started from the first cut.
   */

void Sequencer_Rewind(Sequencer_t *Sequencer,unsigned int Cuts);
  /**< \brief Restarts sequence execution from a previeus cut.
       \param Sequencer    Sequencer reference.
       \param Cuts         Number of cuts to jump back.

       Restarts sequence execution from a previeus cut.
       This function alters cuts order jumping to the start of a previeus cut,
       jump is performed also if the current cut is not ended.
       \p Cuts is the number of cuts to jump back, if \p Cuts is equal to 0
       jump is not performed and sequence is not altered. If \p Cuts is
       greather than the number of cuts in the sequence, the sequence is
       restarted from beginning.
       This function can be called from a sequence cut function.
   */

void Sequencer_FastForward(Sequencer_t *Sequencer,unsigned int Cuts);
  /**< \brief Moves sequence execution to a furter cut.
       \param Sequencer    Sequencer reference.
       \param Cuts         Number of cuts to jump forward.

       Moves sequence execution to a furter cut.
       This function alters cuts order jumping to the start of a furter cut,
       jump is performed also if the current cut is not ended.
       \p Cuts is the number of cuts to jump forward, if \p Cuts is equal to 0
       jump is not performed and sequence is not altered. If \p Cuts is
       greather than the number of cuts in the sequence, the sequence is
       restarted from beginning.
       This function can be called from a sequence cut function.
   */

/***************************************************************************/
#endif

/* $Source$ */

